@@ -5,9 +5,10 @@
 extern "C" {
 #endif
 
-enum { A0_UUID_SIZE = 37 };
+enum { A0_UUID_SIZE = 36 };
 
-typedef char a0_uuid_t[A0_UUID_SIZE];
+// Add one for a null terminator.
+typedef char a0_uuid_t[A0_UUID_SIZE + 1];
 
 void a0_uuidv4(a0_uuid_t out);
 
