@@ -141,12 +141,42 @@ PYBIND11_MODULE(alephzero_bindings, m) {
         return py::bytes(self->payload().data(), self->payload().size());
       })
       .def_property_readonly("payload_view", [](a0::Packet* self) {
-        return py::memoryview::from_memory(self->payload().data(), self->payload().size());
+        return py::memoryview::from_memory((void*)self->payload().data(), self->payload().size(), /* readonly = */ true);
       });
 
   py::implicitly_convertible<py::bytes, a0::Packet>();
   py::implicitly_convertible<py::str, a0::Packet>();
 
+  py::class_<a0::Frame>(m, "Frame")
+      .def_property_readonly("seq", [](const a0::Frame& self) { return self.hdr.seq; })
+      .def_property_readonly("off", [](const a0::Frame& self) { return self.hdr.off; })
+      .def_property_readonly("data", [](const a0::Frame& self) {
+        return py::memoryview::from_memory(self.data, self.hdr.data_size);
+      });
+
+  py::class_<a0::FlatPacket>(m, "FlatPacket")
+      .def(py::init([](const a0::Frame& frame) {
+        // TODO(lshamis): Construct from memoryview instead of Frame.
+        a0::FlatPacket fpkt;
+        fpkt.c = std::make_shared<a0_flat_packet_t>(a0_flat_packet_t{{frame.data, frame.hdr.data_size}});
+        return fpkt;
+      }))
+      .def_property_readonly("id", &a0::FlatPacket::id)
+      .def_property_readonly("headers", [](a0::FlatPacket* self) {
+        std::vector<std::pair<std::string, std::string>> ret;
+        for (size_t i = 0; i < self->num_headers(); i++) {
+          auto hdr = self->header(i);
+          ret.push_back({std::string(hdr.first), std::string(hdr.second)});
+        }
+        return ret;
+      })
+      .def_property_readonly("payload", [](a0::FlatPacket* self) {
+        return py::bytes(self->payload().data(), self->payload().size());
+      })
+      .def_property_readonly("payload_view", [](a0::FlatPacket* self) {
+        return py::memoryview::from_memory((void*)self->payload().data(), self->payload().size(), /* readonly = */ true);
+      });
+
   py::class_<a0::TimeMono>(m, "TimeMono")
       .def_static("now", &a0::TimeMono::now)
       .def_static("parse", &a0::TimeMono::parse)
@@ -175,6 +205,49 @@ PYBIND11_MODULE(alephzero_bindings, m) {
       .def_static("parse", &a0::TimeWall::parse)
       .def("__str__", &a0::TimeWall::to_string);
 
+  py::class_<a0::TransportLocked>(m, "TransportLocked")
+      .def("empty", &a0::TransportLocked::empty)
+      .def("seq_low", &a0::TransportLocked::seq_low)
+      .def("seq_high", &a0::TransportLocked::seq_high)
+      .def("used_space", &a0::TransportLocked::used_space)
+      .def("resize", &a0::TransportLocked::resize)
+      .def("iter_valid", &a0::TransportLocked::iter_valid)
+      .def("frame", &a0::TransportLocked::frame)
+      .def("jump", &a0::TransportLocked::jump)
+      .def("jump_head", &a0::TransportLocked::jump_head)
+      .def("jump_tail", &a0::TransportLocked::jump_tail)
+      .def("has_next", &a0::TransportLocked::has_next)
+      .def("step_next", &a0::TransportLocked::step_next)
+      .def("has_prev", &a0::TransportLocked::has_prev)
+      .def("step_prev", &a0::TransportLocked::step_prev)
+      .def("alloc", &a0::TransportLocked::alloc)
+      .def("alloc_evicts", &a0::TransportLocked::alloc_evicts)
+      .def("commit", &a0::TransportLocked::commit)
+      .def("wait",
+           &a0::TransportLocked::wait,
+           py::call_guard<py::gil_scoped_release>(),
+           py::arg("fn"))
+      .def(
+          "wait",
+          [](a0::TransportLocked* self, std::function<bool()> fn, a0::TimeMono timeout) {
+            return self->wait_until(fn, timeout);
+          },
+          py::call_guard<py::gil_scoped_release>(),
+          py::arg("fn"),
+          py::arg("timeout"))
+      .def(
+          "wait",
+          [](a0::TransportLocked* self, std::function<bool()> fn, double timeout) {
+            return self->wait_for(fn, std::chrono::nanoseconds(uint64_t(1e9 * timeout)));
+          },
+          py::call_guard<py::gil_scoped_release>(),
+          py::arg("fn"),
+          py::arg("timeout"));
+
+  py::class_<a0::Transport>(m, "Transport")
+      .def(py::init<a0::Arena>())
+      .def("lock", &a0::Transport::lock);
+
   py::class_<a0::Middleware>(m, "Middleware");
   m.def("add_time_mono_header", &a0::add_time_mono_header);
   m.def("add_time_wall_header", &a0::add_time_wall_header);
@@ -200,6 +273,31 @@ PYBIND11_MODULE(alephzero_bindings, m) {
       .value("ITER_NEWEST", A0_ITER_NEWEST)
       .export_values();
 
+  py::class_<a0::ReaderSyncZeroCopy>(m, "ReaderSyncZeroCopy")
+      .def(py::init<a0::Arena, a0::ReaderInit, a0::ReaderIter>())
+      .def("can_read", &a0::ReaderSyncZeroCopy::can_read)
+      .def("read", &a0::ReaderSyncZeroCopy::read)
+      .def("read_blocking",
+           py::overload_cast<std::function<void(a0::TransportLocked, a0::FlatPacket)>>(&a0::ReaderSyncZeroCopy::read_blocking),
+           py::call_guard<py::gil_scoped_release>(),
+           py::arg("fn"))
+      .def(
+          "read_blocking",
+          [](a0::ReaderSyncZeroCopy* self, std::function<void(a0::TransportLocked, a0::FlatPacket)> fn, a0::TimeMono timeout) {
+            self->read_blocking(timeout, fn);
+          },
+          py::call_guard<py::gil_scoped_release>(),
+          py::arg("fn"),
+          py::arg("timeout"))
+      .def(
+          "read_blocking",
+          [](a0::ReaderSyncZeroCopy* self, std::function<void(a0::TransportLocked, a0::FlatPacket)> fn, double timeout) {
+            self->read_blocking(a0::TimeMono::now() + std::chrono::nanoseconds(int64_t(timeout * 1e9)), fn);
+          },
+          py::call_guard<py::gil_scoped_release>(),
+          py::arg("fn"),
+          py::arg("timeout"));
+
   py::class_<a0::ReaderSync>(m, "ReaderSync")
       .def(py::init<a0::Arena, a0::ReaderInit, a0::ReaderIter>())
       .def("can_read", &a0::ReaderSync::can_read)
@@ -217,12 +315,20 @@ PYBIND11_MODULE(alephzero_bindings, m) {
           },
           py::call_guard<py::gil_scoped_release>(), py::arg("timeout"));
 
+  py::class_<a0::ReaderZeroCopy, nogil_holder<a0::ReaderZeroCopy>>(m, "ReaderZeroCopy")
+      .def(py::init<a0::Arena,
+                    a0::ReaderInit,
+                    a0::ReaderIter,
+                    std::function<void(a0::TransportLocked, a0::FlatPacket)>>());
+
   py::class_<a0::Reader, nogil_holder<a0::Reader>>(m, "Reader")
       .def(py::init<a0::Arena,
                     a0::ReaderInit,
                     a0::ReaderIter,
                     std::function<void(a0::Packet)>>());
 
+  m.def("read_random_access", &a0::read_random_access);
+
   py::class_<a0::PubSubTopic>(m, "PubSubTopic")
       .def(py::init<std::string, a0::File::Options>(),
            py::arg("name"),
